# FIAIF is an Intelligent firewall, version: $Revision: 1.112 $
#
# description: Automates a packet filtering firewall with iptables.
#
# Script Author:	Anders Fugmann <afu@fugmann.dhs.org>
# 
# FIAIF is an Intelligent firewall
# Copyright (C) 2002-2003 Anders Peter Fugmann
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Please drop me a line and give me your thoughts on this software.
# Patches are more than welcome.

###############################################################################
# Read a zone
# RETURN=0 if reading was sucsessfull 
###############################################################################
function read_zone ()
{
    local ZONE=$1

    # Clear all variables
    for VAR in INPUT OUTPUT FORWARD IP_EXTRA NET_EXTRA MARK ${!REPLY_*} \
	       ${!REDIRECT_*} ${!SNAT_*} ${!LIMIT_*} ${!TC_*} ${!IPSET_*}\
	       IP_DROP MAC_DROP WATCH_IP \
               GLOBAL DHCP_SERVER DYNAMIC;  do
                 unset $VAR
    done

    local ZONE_FILE=CONF_${ZONE}
    ZONE_FILE=${!ZONE_FILE}
    local SOURCE_FILE=${CONF_DIR}/${ZONE_FILE}

    if [[ -f ${SOURCE_FILE} ]]; then       
	source ${SOURCE_FILE}
    else
	if [[ -z "${ZONE_FILE}" ]]; then
	    echo "CONF_${ZONE} variable not set or empty."
	    debug_out "CONF_${ZONE} variable not set or empty."
	else
	    echo "Zonefile not found: ${SOURCE_FILE}"
	    debug_out "Zonefile not found: ${SOURCE_FILE}"
	fi
	let FILE_ERRORS++
	return 1
    fi

    # Accept the old PUBLIC variable.
    if [[ -z "${GLOBAL}" && -n "${PUBLIC}" ]]; then
	GLOBAL=${PUBLIC}
	unset PUBLIC
    fi

    # Accept the old STATIC 
    if [[ -z "${DYNAMIC}" && -n "${STATIC}" ]]; then 
	if (( STATIC == 1 )); then
	    DYNAMIC=0
	else
	    DYNAMIC=1
	fi
	unset STATIC
    fi

    # Default values.
    declare -i GLOBAL=${GLOBAL:=1}
    declare -i DHCP_SERVER=${DHCP_SERVER:=0}
    declare -i DYNAMIC=${DYNAMIC:=1}
    declare -i TC_ENABLE=${TC_ENABLE:=0}
    declare    TC_TYPE=${TC_TYPE:=CBQ}

    # Got all nessesary information about the interface?
    if (( DYNAMIC == 0 )); then	
        # Test that important settings are available.
	if [[ -z "${IP}" ]]; then
	    echo -e "\n${ZONE_FILE}: IP not defined.\n"
	    debug_out "${ZONE_FILE}: IP not defined."
	    return 1
	fi
	if [[ -z "${MASK}" ]]; then
	    echo -e "\n${ZONE_FILE}: MASK not defined.\n"
	    debug_out "${ZONE_FILE}: MASK not defined."
	    return 1
	fi
	if [[ -z "${NET}" ]]; then
	    echo -e "\n${ZONE_FILE}: NET not defined.\n"
	    debug_out "${ZONE_FILE}: NET not defined."
	    return 1
	fi
	if [[ -z "${BCAST}" ]]; then
	    echo -e "\n${ZONE_FILE}: BCAST not defined.\n"
	    debug_out "${ZONE_FILE}: BCAST not defined."
	    return 1
	fi
    fi
}

###############################################################################
# Add reply with
###############################################################################
function add_reply_with ()
{
    local CHAIN=$1
    local PROTOCOL=$2
    local TYPE=$3
    local PORTS=$4

    if [[ -n "${PORTS}" &&
	  ${PORTS} != +([[:alnum:]]):+([[:alnum:]]) && 
	  ${PORTS} != +([[:alnum:]-])*(,+([[:alnum:]-])) ]]; then
	local SINGLE_PORTS=""
	local RANGE_PORTS=""
	for ELEM in $(echo ${PORTS} | sed s/,/\ /g); do
	    if [[ ${ELEM} == +([[:alnum:]]):+([[:alnum:]]) ]]; then
		RANGE_PORTS="${RANGE_PORTS} ${ELEM}"
	    else
		SINGLE_PORTS="${SINGLE_PORTS},${ELEM}"
	    fi
	done
	
	for PORTS in ${SINGLE_PORTS#,*} ${RANGE_PORTS}; do
	    add_reply_with ${CHAIN} ${PROTOCOL} ${TYPE} ${PORTS}
	done
	
	# End processing 
	return
    fi

    
    local PARAM
    case ${PROTOCOL} in
	ICMP | icmp)
	    PARAM="--icmp-type ${PORTS}"
	    ;;
	TCP | tcp | UDP | udp)
	    if [[ ${PORTS} == +([[:alnum:]]):+([[:alnum:]]) ]]; then
		PARAM="--dport ${PORTS}"
	    else
		PARAM="-m multiport --ports ${PORTS}"
	    fi
	    ;;
        *)
	    PARAM=""
	    ;;
    esac
	
    IPTABLES -A ${CHAIN} -p ${PROTOCOL} ${PARAM} \
	-j REJECT --reject-with ${TYPE}
}

###############################################################################
# Limit packet types.
###############################################################################
function add_limit ()
{
    # LIMIT_XXX = <chain> <new_chain> <zone> <policy> <limit> <burst> [protocol [port[,port]*]] 
    local CHAIN=$1
    local NEW_CHAIN=$2
    local ZONE=$3
    local POLICY=$4
    local LIMIT=$5
    local BURST=$6
    local PROTOCOL=$7
    shift 7
    local PORTS=""
    if [[ ${PROTOCOL} == @(TCP|tcp|UDP|udp|ICMP|icmp) ]]; then
	PORTS=$1
	shift
    fi
    local IP=$1
    shift

    if [[ "${IP}" == @(ALL|all) ]]; then 
	print_err ""
	print_err "Error in rules."
	print_err "It seems that a port was specified with protocol 'ALL'."
	print_err "Please recheck INPUT, OUTPUT, FORWARD, MARK and LIMIT rules".
	let RULE_ERRORS++
	IP=$1
	shift
    fi

    check_zone ${ZONE} "Zone ${ZONE} is not defined in LIMIT rule."
    if (( RESULT == 0 )); then 
	let ZONE_ERRORS++
	return
    fi

    # See if the IP is an alias for a defined set.
    local SRC_IP=${IP%*=>*}
    local DST_IP=${IP#*=>*}
    if [[ "${SRC_IP}=>${DST_IP}" == ${IP} ]]; then
	if [[ ${SRC_IP} == *([[:alnum:]]) ]]; then
	    #Expand SRC_IP 
	    local VAR=IPSET_${SRC_IP}
	    for SRC_IP in ${!VAR}; do
		add_limit ${CHAIN} ${NEW_CHAIN} ${ZONE} ${POLICY} ${LIMIT} \
		    ${BURST} ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done
	    
	    #end processing
	    return

	elif [[ ${DST_IP} == *([[:alnum:]]) ]]; then
	    #Expand DST_IP 
	    local VAR=IPSET_${DST_IP}
	    for DST_IP in ${!VAR}; do
		add_limit ${CHAIN} ${NEW_CHAIN} ${ZONE} ${POLICY} ${LIMIT} \
		    ${BURST} ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done

	    #end processing
	    return	    
	fi
    fi

    # Split the ports if nessesary
    if [[ -n "${PORTS}" && 
	   ${PORTS} != +([[:alnum:]]):+([[:alnum:]]) && 
	   ${PORTS} != +([[:alnum:]-])*(,+([[:alnum:]-])) ]]; then
	local SINGLE_PORTS=""
	local RANGE_PORTS=""
	for ELEM in $(echo ${PORTS} | sed s/,/\ /g); do
	    if [[ ${ELEM} == +([[:alnum:]]):+([[:alnum:]]) ]]; then
		RANGE_PORTS="${RANGE_PORTS} ${ELEM}"
	    else
		SINGLE_PORTS="${SINGLE_PORTS},${ELEM}"
	    fi
	done

	for PORTS in ${SINGLE_PORTS#,*} ${RANGE_PORTS}; do
	    add_limit ${CHAIN} ${NEW_CHAIN} ${ZONE} ${POLICY} ${LIMIT} \
		${BURST} ${PROTOCOL} ${PORTS} ${IP}
	done
 
	return
    fi
    

    local PARAM
    get_protocol_port_ip IN ${PROTOCOL} ${PORTS} ${IP}
    PARAM="${PARAM} ${RETURN}"

    local LIMIT="-m limit --limit ${LIMIT} --limit-burst ${BURST}" 
    IPTABLES -N ${NEW_CHAIN} 
    IPTABLES -A ${NEW_CHAIN} ${LIMIT} -j RETURN
    IPTABLES -A ${NEW_CHAIN} -j LOG_LIMIT_${POLICY}
    IPTABLES -A ${CHAIN} ${PARAM} -j ${NEW_CHAIN}
}

###############################################################################
# Mark packets.
# <chain_id> <chain_pre> <chain_out> <zone|ALL> <mark> <protocol|all> <ports|all> <ipmask=>ipmask>
###############################################################################
function add_mark ()
{
    local RULE_ID=$1
    local CHAIN_PRE=$2
    local CHAIN_OUT=$3
    local SRC_ZONE=$4
    local MARK=$5
    local PROTOCOL=$6
    shift 6
    local PORTS=""
    if [[ ${PROTOCOL} == @(TCP|tcp|UDP|udp|ICMP|icmp) ]]; then
	PORTS=$1
	shift
    fi
    local IP=$1
    shift

    if [[ "${IP}" == @(ALL|all) ]]; then 
	print_err ""
	print_err "Error in rules."
	print_err "It seems that a port was specified with protocol 'ALL'."
	print_err "Please recheck INPUT, OUTPUT, FORWARD, MARK and LIMIT rules".
	let RULE_ERRORS++
	shift
	IP=$1
    fi
	

    check_zone ${ZONE} "Zone ${ZONE} is not defined in MARK rule."
    if (( RESULT ==  0 )); then 
	let ZONE_ERRORS++
	return
    fi

    # See if the IP is an alias for a defined set.
    local SRC_IP=${IP%*=>*}
    local DST_IP=${IP#*=>*}
    if [[ "${SRC_IP}=>${DST_IP}" == ${IP} ]]; then
	if [[ ${SRC_IP} == *([[:alnum:]]) ]]; then
	    #Expand SRC_IP 
	    local VAR=IPSET_${SRC_IP}
	    for SRC_IP in ${!VAR}; do
		add_mark ${RULE_ID} ${CHAIN_PRE} ${CHAIN_OUT} ${SRC_ZONE} \
		    ${MARK} ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done
	    
	    #end processing
	    return

	elif [[ ${DST_IP} == *([[:alnum:]]) ]]; then
	    #Expand DST_IP 
	    local VAR=IPSET_${DST_IP}
	    for DST_IP in ${!VAR}; do
		add_mark ${RULE_ID} ${CHAIN_PRE} ${CHAIN_OUT} ${SRC_ZONE} \
		    ${MARK} ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done

	    #end processing
	    return	    
	fi
    fi

    # Split the ports if nessesary
    if [[ -n "${PORTS}" && 
	   ${PORTS} != +([[:alnum:]]):+([[:alnum:]]) && 
	   ${PORTS} != +([[:alnum:]-])*(,+([[:alnum:]-])) ]]; then
	local SINGLE_PORTS=""
	local RANGE_PORTS=""
	for ELEM in $(echo ${PORTS} | sed s/,/\ /g); do
	    if [[ ${ELEM} == +([[:alnum:]]):+([[:alnum:]]) ]]; then
		RANGE_PORTS="${RANGE_PORTS} ${ELEM}"
	    else
		SINGLE_PORTS="${SINGLE_PORTS},${ELEM}"
	    fi
	done

	for PORTS in ${SINGLE_PORTS#,*} ${RANGE_PORTS}; do
	    add_mark ${RULE_ID} ${CHAIN_PRE} ${CHAIN_OUT} ${SRC_ZONE} \
		${MARK} ${PROTOCOL} ${PORTS} ${IP} 
	done
 
	return
    fi

    local PARAM 
	
    get_protocol_port_ip IN ${PROTOCOL} ${PORTS} ${IP}
    PARAM=${RETURN}
    
    # Get the device
    if [[ "${SRC_ZONE}" == "ALL" ]]; then
	IPTABLES -t mangle -A ${CHAIN_PRE} ${PARAM} -j MARK --set-mark ${MARK}
	IPTABLES -t mangle -A ${CHAIN_OUT} ${PARAM} -j MARK --set-mark ${MARK}
    elif [[ "${SRC_ZONE}" == "${ZONE}" ]]; then
	IPTABLES -t mangle -A ${CHAIN_OUT} ${PARAM} -j MARK --set-mark ${MARK}
    else
	local DEVICE=${ZONE}_DEV
	DEVICE=${!DEVICE}
	local GLOBAL=${ZONE}_GLOBAL
	GLOBAL=${!GLOBAL}

	if (( GLOBAL == 1 || DYNAMIC == 1)); then
	    IPTABLES -t mangle -A ${CHAIN_PRE} -i ${DEVICE} \
		${PARAM} -j MARK --set-mark ${MARK}
	else
	    # Need to use the source of the packet. 
	    # We create a new CHAIN, and forward to it, if the source matches.
	    # We also need a unique, explanatory name for the chain.
	    local MARK_CHAIN=MARK_${RULE_ID}
	    IPTABLES -t mangle -N ${MARK_CHAIN}
	    IPTABLES -t mangle -A ${MARK_CHAIN} ${PARAM} \
		-j MARK --set-mark ${MARK}
	    
	    local ZONE_NETS=${ZONE}_NETS
	    for NET in ${!ZONE_NETS}; do
		IPTABLES -t mangle -A ${CHAIN_PRE} -i ${DEVICE} --src ${NET} \
		    -j ${MARK_CHAIN}
	    done
	fi
    fi
}


###############################################################################
# Add redirect rules.
###############################################################################
function add_redirect ()
{
    local CHAIN=$1
    local SNAT_CHAIN=$2
    local FORWARD_CHAIN=$3
    local RULE_ID=$4
    local PROTOCOL=$5
    local DEST=$6
    local DNAT=$7
    
    local PARAM=""

    # Lets split up the old destination
    if [[ ${DEST} == *:* ]]; then
	local DEST_ADDR=${DEST%:*}
        local DEST_PORT=${DEST#*:}
    else
	local DEST_ADDR=${DEST}
        local DEST_PORT=""
    fi	

    if [[ -n "${DEST_ADDR}" ]]; then
	PARAM="-d ${DEST_ADDR}"
    fi

    if [[ -n "${DEST_PORT}" ]]; then
	local START_PORT=${DEST_PORT%-*}
	local END_PORT=${DEST_PORT#*-}

	local PORT_RANGE=""
	if [[ "${END_PORT}" != "${START_PORT}" ]]; then
	    PORT_RANGE="--dport ${START_PORT}:${END_PORT}"
	else
	    PORT_RANGE="--dport ${START_PORT}"
	fi

	PARAM="${PARAM} ${PORT_RANGE}" 
    fi

    local REDIRECT
    # Test if destination is localhost.
    if [[ ${DNAT} == 127.0.0.1* || ${DNAT} == localhost* || \
	  ${DNAT} = :* ]]; then
	REDIRECT="-j REDIRECT"
	# Tests if there are ports involved
	if [[ ${DNAT} == *:* ]]; then 
	    REDIRECT="${REDIRECT} --to-ports ${DNAT#*:}"
	fi
    else
	REDIRECT="-j DNAT --to-destination ${DNAT}"
	# Test if a SNAT is needed to route back to the same network.
	if (( DYNAMIC == 0 && GLOBAL == 0 )); then
	    if [[ ${DNAT} == *:* ]]; then
		local NEW_IP=${DNAT%:*}
		local NEW_PORT=${DNAT#*:}
	    else
		local NEW_IP=${DNAT}
		local NEW_PORT=""
	    fi		
	    # Need to examine all networks - A zone can have NET_EXTRA
	    local NETWORKS=${ZONE}_NETS
	    NETWORKS=${!NETWORKS}
	    local RET NET
	    RET=0
	    for NET in ${NETWORKS}; do
		if ip_in_network ${NET%/*} ${NET#*/} ${NEW_IP}; then
		    RET=1
		    break;
		fi
	    done
	    if (( RET == 1 )); then
		# Add SNAT for packets going back to the zone.
		local SNAT_PARAM
		SNAT_PARAM="-p ${PROTOCOL} -d ${NEW_IP}"
		if [[ -n "${NEW_PORT}" ]]; then
		    SNAT_PARAM="${SNAT_PARAM} --dport ${NEW_PORT}"
		fi
		local SNAT_ADDR
		if [[ -n "${DEST_ADDR}" ]];then
		    SNAT_ADDR=${DEST_ADDR}
		else
		    SNAT_ADDR=${IP}
		fi
		
		local NEW_SNAT_CHAIN=${RULE_ID}_${NAME}
		local NEW_FORWARD_CHAIN=${RULE_ID}_${NAME}
		IPTABLES -t nat -N ${NEW_SNAT_CHAIN}
		IPTABLES -t nat -A ${NEW_SNAT_CHAIN} ${SNAT_PARAM} \
		    -j SNAT --to-source ${SNAT_ADDR}

		IPTABLES -N ${NEW_FORWARD_CHAIN}
		IPTABLES -A ${NEW_FORWARD_CHAIN} ${SNAT_PARAM} -i ${DEV} \
		    -j ACCEPT

		for NET in ${NETWORKS}; do
		    IPTABLES -t nat -A ${SNAT_CHAIN} -s ${NET} -j ${NEW_SNAT_CHAIN}
		    IPTABLES        -A ${FORWARD_CHAIN} -s ${NET} -j ${NEW_FORWARD_CHAIN}
		done
	    fi
	fi
    fi
    IPTABLES -t nat -A ${CHAIN} -p ${PROTOCOL} ${PARAM} ${REDIRECT}  
}

###############################################################################
# add_snat: <zone> [ports [ip[/mask]]]
###############################################################################
function add_snat ()
{
    local RULE_ID=$1
    local ZONE=$2
    local PORTS=$3
    local MASK=$4
    

    local NAT_TYPE=""
    local PARAM=""
    local PARAM_PORTS=""
    local SOURCE=""
    
    check_zone ${ZONE} "Zone ${ZONE} is not defined in SNAT rule."
    if (( RESULT == 0 )); then 
	let ZONE_ERRORS++
	return
    fi
    local SNAT_CHAIN=POSTROUTING_NAT_${ZONE}
    local ZONE_DEV=${ZONE}_DEV
    local ZONE_IP=${ZONE}_IP
    local ZONE_DYNAMIC=${ZONE}_DYNAMIC
    local ZONE_IPS=${ZONE}_IPS

    ZONE_DEV=${!ZONE_DEV}
    ZONE_IP=${!ZONE_IP}
    ZONE_DYNAMIC=${!ZONE_DYNAMIC}
    ZONE_IPS=${!ZONE_IPS}

    local NEW_SNAT_CHAIN=${RULE_ID}_${NAME}
    IPTABLES -t nat -N ${NEW_SNAT_CHAIN}

    local NETWORKS=${NAME}_NETS
    NETWORKS=${!NETWORKS}
    local NET
    if (( GLOBAL == 0 && DYNAMIC == 0 )); then
	for NET in ${NETWORKS}; do
	    IPTABLES -t nat -A ${SNAT_CHAIN} -s ${NET} -j ${NEW_SNAT_CHAIN}
	done
    else
	# No way to test. SNAT all
	IPTABLES -A -t nat ${SNAT_CHAIN} -j ${NEW_SNAT_CHAIN}
    fi
	
    
    # Always use port specification. 
    # If none is defined, then use the setting from 
    # /proc/sys/net/ipv4/ip_local_port_range.

    if [[ -z "${PORTS}" && -f /proc/sys/net/ipv4/ip_local_port_range ]]; then
     	PORTS="$(cut -f1 /proc/sys/net/ipv4/ip_local_port_range)"
     	PORTS="${PORTS}-$(cut -f2 /proc/sys/net/ipv4/ip_local_port_range)"
    fi

    if (( ZONE_DYNAMIC == 0 )); then
	NAT_TYPE=SNAT
	local IP
	for IP in ${ZONE_IPS}; do
	    PARAM="${PARAM} --to-source ${IP}"
	    PARAM_PORTS="${PARAM_PORTS} --to-source ${IP}:${PORTS}"
	done
    else
	NAT_TYPE=MASQUERADE
	PARAM=""
	PARAM_PORTS="--to-ports ${PORTS}"
    fi

    if [[ -n "${MASK}" ]]; then
	SOURCE="-s ${MASK}"
    fi
    
    for PROTOCOL in TCP UDP; do
	IPTABLES -t nat -A ${SNAT_CHAIN} ${SOURCE} -p ${PROTOCOL} \
	    -j ${NAT_TYPE} ${PARAM_PORTS}
    done
    IPTABLES -t nat -A ${SNAT_CHAIN} ${SOURCE} \
	-j ${NAT_TYPE} ${PARAM}
}

###############################################################################
# Add deafult rules for in/out.
###############################################################################
function add_zone_rule()
{	
    local DIRECTION=$1
    local CHAIN=$2
    local POLICY=$3
    local PROTOCOL=$4
    shift 4
    local PORTS=""
    if [[ ${PROTOCOL} == @(TCP|tcp|UDP|udp|ICMP|icmp) ]]; then
	PORTS=$1
	shift
    fi
    local IP=$1
    shift

    if [[ "${IP}" == @(ALL|all) ]]; then 
	print_err ""
	print_err "Error in rules."
	print_err "It seems that a port was specified with protocol 'ALL'."
	print_err "Please recheck INPUT, OUTPUT, FORWARD, MARK and LIMIT rules".
	let RULE_ERRORS++
	IP=$1
	shift
    fi

    # See if the IP is an alias for a defined set.
    if [[ "${IP}" == +(?)=\>+(?) ]]; then
	local SRC_IP=${IP/=>*}
	local DST_IP=${IP//*=>}
	if [[ ${SRC_IP} == *([[:alnum:]]) ]]; then
	    #Expand SRC_IP 
	    local VAR=IPSET_${SRC_IP}
	    for SRC_IP in ${!VAR}; do
		add_zone_rule ${DIRECTION} ${CHAIN} ${POLICY} \
		    ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done
	    
	    #end processing
	    return

	elif [[ ${DST_IP} == *([[:alnum:]]) ]]; then
	    #Expand DST_IP 
	    local VAR=IPSET_${DST_IP}
	    for DST_IP in ${!VAR}; do
		add_zone_rule ${DIRECTION} ${CHAIN} ${POLICY} \
		    ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done

	    #end processing
	    return	    
	fi
    fi
	
    # Split the ports if nessesary
    
    if [[ -n "${PORTS}" &&  
	   ${PORTS} != +([[:alnum:]]):+([[:alnum:]]) && 
	   ${PORTS} != +([[:alnum:]-])*(,+([[:alnum:]-])) ]]; then
	local SINGLE_PORTS=""
	local RANGE_PORTS=""
	for ELEM in $(echo ${PORTS} | sed s/,/\ /g); do
	    if [[ ${ELEM} == +([[:alnum:]]):+([[:alnum:]]) ]]; then
		RANGE_PORTS="${RANGE_PORTS} ${ELEM}"
	    else
		SINGLE_PORTS="${SINGLE_PORTS},${ELEM}"
	    fi
	done

	for PORTS in ${SINGLE_PORTS#,*} ${RANGE_PORTS}; do
	    add_zone_rule ${DIRECTION} ${CHAIN} ${POLICY} ${PROTOCOL} \
		${PORTS} ${IP}	    
	done
 
	return
    fi

    # Actual function
    get_protocol_port_ip ${DIRECTION} ${PROTOCOL} ${PORTS} ${IP} 
    IPTABLES -A ${CHAIN} ${RETURN} -j LOG_${POLICY}
}    

###############################################################################
# add_zone_forward: chain zone protocol ports policy
###############################################################################
function add_zone_forward ()
{
    local RULE_ID=$1
    local CHAIN=$2
    local ZONE=$3
    local POLICY=$4
    local PROTOCOL=$5
    shift 5
    local PORTS=""
    if [[ ${PROTOCOL} == @(TCP|tcp|UDP|udp|ICMP|icmp) ]]; then
	PORTS=$1
	shift
    fi
    local IP=$1
    shift

    if [[ "${IP}" == @(ALL|all) ]]; then 
	print_err ""
	print_err "Error in rules."
	print_err "It seems that a port was specified with protocol 'ALL'."
	print_err "Please recheck INPUT, OUTPUT, FORWARD, MARK and LIMIT rules".
	let RULE_ERRORS++
	IP=$1
	shift
    fi

    check_zone ${ZONE} "Zone ${ZONE} is not defined in FORWARD rule."
    if (( RESULT == 0 )); then 
	let ZONE_ERRORS++
	return
    fi

    # See if the IP is an alias for a defined set.
    local SRC_IP=${IP%*=>*}
    local DST_IP=${IP#*=>*}
    if [[ "${SRC_IP}=>${DST_IP}" == ${IP} ]]; then
	if [[ ${SRC_IP} == *([[:alnum:]]) ]]; then
	    #Expand SRC_IP 
	    local VAR=IPSET_${SRC_IP}
	    for SRC_IP in ${!VAR}; do
		add_zone_forward ${RULE_ID} ${CHAIN} ${ZONE} \
		    ${POLICY} ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done
	    
	    #end processing
	    return

	elif [[ ${DST_IP} == *([[:alnum:]]) ]]; then
	    #Expand DST_IP 
	    local VAR=IPSET_${DST_IP}
	    for DST_IP in ${!VAR}; do
		add_zone_forward ${RULE_ID} ${CHAIN} ${ZONE} \
		    ${POLICY} ${PROTOCOL} ${PORTS} ${SRC_IP}=\>${DST_IP}
	    done

	    #end processing
	    return	    
	fi
    fi

    # Test if the source is set. If not then use the zone source.
    # Else disregard it.

    # First split the ports.
    # Split the ports if nessesary
    if [[ -n "${PORTS}" && 
	   ${PORTS} != +([[:alnum:]]):+([[:alnum:]]) && 
	   ${PORTS} != +([[:alnum:]-])*(,+([[:alnum:]-])) ]]; then
	local SINGLE_PORTS=""
	local RANGE_PORTS=""
	for ELEM in $(echo ${PORTS} | sed s/,/\ /g); do
	    if [[ ${ELEM} == +([[:alnum:]]):+([[:alnum:]]) ]]; then
		RANGE_PORTS="${RANGE_PORTS} ${ELEM}"
	    else
		SINGLE_PORTS="${SINGLE_PORTS},${ELEM}"
	    fi
	done

	for PORTS in ${SINGLE_PORTS#,*} ${RANGE_PORTS} ; do
	    add_zone_forward ${RULE_ID} ${CHAIN} ${ZONE} ${POLICY} \
		${PROTOCOL} ${PORTS} ${IP}
	done
 
	return
    fi

    local PARAM 
    get_protocol_port_ip IN ${PROTOCOL} ${PORTS} ${IP}
    PARAM=${RETURN}
    
    # Get the device
    if [[ "${ZONE}" == @(ALL|all) ]]; then
	# Test the source of the packet.
	IPTABLES -A ${CHAIN} ${PARAM} -j LOG_${POLICY}
    else
	local DEVICE=${ZONE}_DEV
	DEVICE=${!DEVICE}
	local GLOBAL=${ZONE}_GLOBAL
	GLOBAL=${!GLOBAL}

	if (( GLOBAL == 1 || DYNAMIC == 1 )); then
	    IPTABLES -A ${CHAIN} -i ${DEVICE} ${PARAM} -j LOG_${POLICY}
	else
	    # Need to use the source of the packet. 
	    # We create a new CHAIN, and forward to it, if the source matches.
	    # We also need a unique, explanatory name for the chain.
	    local FORWARD_CHAIN=FORWARD_${RULE_ID}
	    IPTABLES -N ${FORWARD_CHAIN}
	    IPTABLES -A ${FORWARD_CHAIN} ${PARAM} -j LOG_${POLICY}
	    
	    local ZONE_NETS=${ZONE}_NETS
	    for NET in ${!ZONE_NETS}; do
		IPTABLES -A ${CHAIN} -i ${DEVICE} --src ${NET} \
		    -j ${FORWARD_CHAIN}
	    done
	fi
    fi
}

###############################################################################
# Setup Zone chains.
# Give zone name as argument
###############################################################################
function zone_chains ()
{
    local ZONE=$1
    debug_out "Creating global chains for zone: ${ZONE}"
 
    # FILTER chains
    for CHAIN in ${BUILT_IN_CHAINS_filter} SEND; do
	IPTABLES -N ${CHAIN}_${ZONE}
    done
    for CHAIN in ${BUILT_IN_CHAINS_mangle}; do
	IPTABLES -t mangle -N ${CHAIN}_MANGLE_${ZONE}
    done

    for CHAIN in ${BUILT_IN_CHAINS_nat}; do
	IPTABLES -t nat -N ${CHAIN}_NAT_${ZONE}
    done
    
}

###############################################################################
# Setup zone specific rules.
###############################################################################
function configure_zone ()
{
    local ZONE=$1
    read_zone ${ZONE}

    # Stop processing if the device is not available due 
    # to a configuration error.
    local ZONE_AVAILABLE=${ZONE}_AVAILABLE
    ZONE_AVAILABLE=${!ZONE_AVAILABLE}
    if (( ZONE_AVAILABLE == 0 )); then
	return
    fi
    debug_out ""
    debug_out ""
    debug_out "Configuring zone: ${NAME}"
    debug_out "DEV=${DEV}, DYNAMIC=${DYNAMIC}, GLOBAL=${GLOBAL}"
    if (( DYNAMIC == 0 )); then
	debug_out "IP=${IP}, MASK=${MASK}, NET=${NET}, BCAST=${BCAST}"
    fi
    debug_out ""
    debug_out "Creating zone chains"

    local PRE_NAT_CHAIN=PREROUTING_NAT_${NAME}
    local POST_NAT_CHAIN=POSTROUTING_NAT_${NAME}

    local PRE_MANGLE_CHAIN=PREROUTING_MANGLE_${NAME}
    local POST_MANGLE_CHAIN=POSTROUTING_MANGLE_${NAME}
    local OUTPUT_MANGLE_CHAIN=OUTPUT_MANGLE_${NAME}

    # Allow server to answer DHCP queries.
    debug_out "DHCP_SERVER=${DHCP_SERVER}"
    if (( DHCP_SERVER == 1 )); then 
	IPTABLES -A DEV_${DEV}_SRC -p UDP -s 0.0.0.0 -d 255.255.255.255 \
	    --dport bootps -j LOG_ACCEPT
	IPTABLES -A DEV_${DEV}_DST -p UDP -d 255.255.255.255 \
	    --dport bootpc -j LOG_ACCEPT 
    fi
    
    # Allow server to send DHCP queries.
    debug_out "DYNAMIC=${DYNAMIC}"
    if (( DYNAMIC == 1 )); then 
	IPTABLES -A DEV_${DEV}_SRC -p UDP -d 255.255.255.255 \
	    --dport bootpc -j LOG_ACCEPT
	IPTABLES -A DEV_${DEV}_DST -p UDP -d 255.255.255.255 -s 0.0.0.0 \
	    --dport bootps -j LOG_ACCEPT 
    fi
 
    # Forward packets to queues.    
    # Only allow packets that can originate from the device. 
    # This way we can have more than one zone on one interface
    # - Unless its global, then there can be only one zone per interface.

    debug_out "GLOBAL=${GLOBAL}"

    local CHAIN_INPUT=INPUT_${NAME}
    local CHAIN_OUTPUT=OUTPUT_${NAME}
    local CHAIN_FORWARD=FORWARD_${NAME}
    local CHAIN_SEND=SEND_${NAME}

    # Both GLOBAL and NONGLOBAL have net and net_extra specified. 
    if (( DYNAMIC == 0 )); then
	local NETWORK
	for NETWORK in ${NET} ${NET_EXTRA}; do
	    IPTABLES -A DEV_${DEV}_DST -d ${NETWORK} -j RETURN
	    IPTABLES -A DEV_${DEV}_SRC -s ${NETWORK} -j RETURN
	done
    fi


    if (( GLOBAL == 1 || DYNAMIC == 1 )); then	
	
	IPTABLES -t nat -A PREROUTING -i ${DEV} -j ${PRE_NAT_CHAIN}
	IPTABLES -t nat -A POSTROUTING -o ${DEV} -j ${POST_NAT_CHAIN}
	
	IPTABLES -t mangle -A PREROUTING -i ${DEV} -j ${PRE_MANGLE_CHAIN}
	IPTABLES -t mangle -A POSTROUTING -o ${DEV} -j ${POST_MANGLE_CHAIN}
	IPTABLES -t mangle -A OUTPUT -o ${DEV} -j ${OUTPUT_MANGLE_CHAIN}

	if (( DYNAMIC == 0 )); then
	    IPTABLES -A DEV_${DEV}_DST -j PRIVATE_DST
	    IPTABLES -A DEV_${DEV}_SRC -j PRIVATE_SRC
	fi

	IPTABLES -A DEV_${DEV}_DST -j RESERVED_DST
	IPTABLES -A DEV_${DEV}_DST -j RETURN	

	IPTABLES -A DEV_${DEV}_SRC -j RESERVED_SRC
	IPTABLES -A DEV_${DEV}_SRC -j RETURN	

	# Jump to corect chains.
	IPTABLES -A INPUT_NEW_${DEV} -j ${CHAIN_INPUT}
	IPTABLES -A OUTPUT_NEW_${DEV} -j ${CHAIN_OUTPUT}
	IPTABLES -A FORWARD_NEW_${DEV} -j ${CHAIN_FORWARD}
	IPTABLES -A SEND_NEW_${DEV} -j ${CHAIN_SEND}

    else
	for NETWORK in ${NET} ${NET_EXTRA}; do

	    IPTABLES -t nat -A PREROUTING -i ${DEV} -s ${NETWORK} \
		-j ${PRE_NAT_CHAIN}
	    IPTABLES -t nat -A POSTROUTING -o ${DEV} -d ${NETWORK} \
		-j ${POST_NAT_CHAIN}

	    IPTABLES -t mangle -A PREROUTING -i ${DEV} -s ${NETWORK} \
		-j ${PRE_MANGLE_CHAIN}
	    IPTABLES -t mangle -A POSTROUTING -o ${DEV} -d ${NETWORK} \
		-j ${POST_MANGLE_CHAIN}
	    IPTABLES -t mangle -A OUTPUT -o ${DEV} -d ${NETWORK} \
		-j ${OUTPUT_MANGLE_CHAIN}

	    IPTABLES -A INPUT_NEW_${DEV} -s ${NETWORK} -j ${CHAIN_INPUT}
	    IPTABLES -A OUTPUT_NEW_${DEV} -d ${NETWORK} -j ${CHAIN_OUTPUT}
	    IPTABLES -A FORWARD_NEW_${DEV} -d ${NETWORK} -j ${CHAIN_FORWARD}
	    IPTABLES -A SEND_NEW_${DEV} -s ${NETWORK} -j ${CHAIN_SEND}
	done
    fi


    # Setup redirects. These are added to the NAT chain.
    local REDIRECT
    for REDIRECT in ${!REDIRECT_*}; do
	debug_out "${REDIRECT}=${!REDIRECT}"
	add_redirect ${PRE_NAT_CHAIN} ${POST_NAT_CHAIN} ${CHAIN_FORWARD} \
	    ${REDIRECT} ${!REDIRECT} 
    done   
    
    local I
    for ((I=0;I<${#MARK[*]};I++)); do
 	debug_out "MARK[${I}]=${MARK[I]}"   
	add_mark ${ZONE}${I} ${PRE_MANGLE_CHAIN} ${OUTPUT_MANGLE_CHAIN} \
	    ${MARK[I]}
    done   

    local VAR
    #Setup Watch rules:
    for VAR in ${WATCH_IP}; do
	debug_out "WATCH_IP: ${VAR}"
	IPTABLES -t mangle -A ${PRE_MANGLE_CHAIN} -s ${VAR} -j LOG_WATCH
	IPTABLES -t mangle -A ${POST_MANGLE_CHAIN} -d ${VAR} -j LOG_WATCH
    done

    # Setup drop rules
    for VAR in ${MAC_DROP}; do
	debug_out "MAC_DROP: ${VAR}"
	IPTABLES -A ${CHAIN_INPUT}  -m mac --mac-source ${VAR} -j LOG_DROP
	IPTABLES -A ${CHAIN_SEND} -m mac --mac-source ${VAR} -j LOG_DROP   
    done
    
    for VAR in ${IP_DROP}; do
	debug_out "IP_DROP: ${VAR}"
	IPTABLES -A ${CHAIN_INPUT}  -s ${VAR} -j LOG_DROP
	IPTABLES -A ${CHAIN_OUTPUT} -d ${VAR} -j LOG_DROP
	IPTABLES -A ${CHAIN_FORWARD} -d ${VAR} -j LOG_DROP   
	IPTABLES -A ${CHAIN_SEND} -s ${VAR} -j LOG_DROP
    done
    
    # Source NAT are done in the NAT table - postrouting.
    for VAR in ${!SNAT_*}; do
	debug_out "${VAR}: ${!VAR}"
	add_snat ${VAR} ${!VAR} 
    done
       
    # Setup reply_with
    local REPLY
    for REPLY in ${!REPLY_*}; do
	debug_out "${REPLY}=${!REPLY}"
	add_reply_with ${CHAIN_INPUT} ${!REPLY}
    done

    # Limit number of packets.
    local LIMIT
    for LIMIT in ${!LIMIT_*}; do
	debug_out "${LIMIT}=${!LIMIT}"
	if [[ "${!LIMIT%% *}" = "${NAME}" ]]; then
	    add_limit ${CHAIN_INPUT} ${LIMIT}_${ZONE} ${!LIMIT}
	else
	    add_limit ${CHAIN_FORWARD} ${LIMIT}_${ZONE} ${!LIMIT}
	fi
    done

    # Create rules for INPUT.
    if (( DHCP_SERVER == 1 )); then
	# Allow renewal of DHCP leases.
        IPTABLES -A ${CHAIN_INPUT} -p udp \
	    --sport bootpc --dport bootps -j ACCEPT
    fi

    # Add user chains
    local CHAIN
    local FIAIF_CHAIN
    for CHAIN in INPUT OUTPUT FORWARD; do
	IPTABLES -N USER_${CHAIN}_${NAME}
	FIAIF_CHAIN=CHAIN_${CHAIN}
	IPTABLES -I ${!FIAIF_CHAIN} -j USER_${CHAIN}_${NAME}
    done

    for (( I=0;I<${#INPUT[*]};I++ )); do
 	debug_out "INPUT[${I}]=${INPUT[I]}"   
	add_zone_rule IN ${CHAIN_INPUT} ${INPUT[I]}
    done

    # Create rules for OUTPUT.
    for (( I=0;I<${#OUTPUT[*]};I++ )); do
 	debug_out "OUTPUT[${I}]=${OUTPUT[I]}"   
	add_zone_rule OUT ${CHAIN_OUTPUT} ${OUTPUT[I]}
    done
    
    # Create rules for FORWARD.
    for (( I=0;I<${#FORWARD[*]};I++ )); do
 	debug_out "FORWARD[${I}]=${FORWARD[I]}"   
	add_zone_forward ${ZONE}${I} ${CHAIN_FORWARD} ${FORWARD[I]}
    done

    debug_out "Log all unmatched packets in this zone"
    IPTABLES -A ${CHAIN_INPUT} -j LOG_MISS
    IPTABLES -A ${CHAIN_OUTPUT} -j LOG_MISS
    IPTABLES -A ${CHAIN_FORWARD} -j LOG_MISS

    debug_out "Done configuring zone"
}

